@@ -1,12 +1,14 @@
 """
 aurora-data-api - A Python DB-API 2.0 client for the AWS Aurora Serverless Data API
 """
-import os, datetime, ipaddress, uuid, time, random, string, logging, itertools, reprlib
+import os, datetime, ipaddress, time, random, string, logging, itertools, reprlib, json, re
+from uuid import UUID
 from decimal import Decimal
 from collections import namedtuple
 from .exceptions import (Warning, Error, InterfaceError, DatabaseError, DataError, OperationalError, IntegrityError,
                          InternalError, ProgrammingError, NotSupportedError)
 from .mysql_error_codes import MySQLErrorCodes
+from .postgresql_error_codes import PostgreSQLErrorCodes
 import boto3
 
 apilevel = "2.0"
@@ -34,9 +36,12 @@ ColumnDescription.__new__.__defaults__ = (None,) * len(ColumnDescription._fields
 
 logger = logging.getLogger(__name__)
 
+postgresql_error_reg = re.compile(r'^ERROR: (.*)[\s+]+Position: ([0-9]+); SQLState: ([0-9A-Z]+)$')
+
 
 class AuroraDataAPIClient:
-    def __init__(self, dbname=None, aurora_cluster_arn=None, secret_arn=None, rds_data_client=None, charset=None):
+    def __init__(self, dbname=None, aurora_cluster_arn=None, secret_arn=None, rds_data_client=None, charset=None,
+                 continue_after_timeout=None):
         self._client = rds_data_client
         if rds_data_client is None:
             self._client = boto3.client("rds-data")
@@ -45,6 +50,7 @@ class AuroraDataAPIClient:
         self._secret_arn = secret_arn or os.environ.get("AURORA_SECRET_ARN")
         self._charset = charset
         self._transaction_id = None
+        self._continue_after_timeout = continue_after_timeout
 
     def close(self):
         pass
@@ -76,7 +82,8 @@ class AuroraDataAPIClient:
                                      dbname=self._dbname,
                                      aurora_cluster_arn=self._aurora_cluster_arn,
                                      secret_arn=self._secret_arn,
-                                     transaction_id=self._transaction_id)
+                                     transaction_id=self._transaction_id,
+                                     continue_after_timeout=self._continue_after_timeout)
         if self._charset:
             cursor.execute("SET character_set_client = '{}'".format(self._charset))
         return cursor
@@ -110,13 +117,13 @@ class AuroraDataAPICursor:
         "cidr": ipaddress.ip_network,
         "date": datetime.date,
         "inet": ipaddress.ip_address,
-        "json": dict,  # TODO
+        "json": "json",
         "jsonb": dict,  # TODO
         "money": str,  # TODO
         "text": str,
         "time": datetime.time,
         "timestamp": datetime.datetime,
-        "uuid": uuid.uuid4,
+        "uuid": UUID,
         "numeric": Decimal,
         "decimal": Decimal
     }
@@ -133,10 +140,13 @@ class AuroraDataAPICursor:
         datetime.date: "DATE",
         datetime.time: "TIME",
         datetime.datetime: "TIMESTAMP",
-        Decimal: "DECIMAL"
+        Decimal: "DECIMAL",
+        "json": "JSON",
+        UUID: "UUID"
     }
 
-    def __init__(self, client=None, dbname=None, aurora_cluster_arn=None, secret_arn=None, transaction_id=None):
+    def __init__(self, client=None, dbname=None, aurora_cluster_arn=None, secret_arn=None, transaction_id=None,
+                 continue_after_timeout=None):
         self.arraysize = 1000
         self.description = None
         self._client = client
@@ -147,6 +157,7 @@ class AuroraDataAPICursor:
         self._current_response = None
         self._iterator = None
         self._paging_state = None
+        self._continue_after_timeout = continue_after_timeout
 
     def prepare_param(self, param_name, param_value):
         if param_value is None:
@@ -202,19 +213,25 @@ class AuroraDataAPICursor:
     def _format_parameter_set(self, parameters):
         return [self.prepare_param(k, v) for k, v in parameters.items()]
 
-    def _get_database_error(self, original_error):
-        # TODO: avoid SHOW ERRORS if on postgres (it's a useless network roundtrip)
+    def _get_database_error(self, origin_error):
+        error_msg = getattr(origin_error, "response", {}).get("Error", {}).get("Message", "")
         try:
-            err_res = self._client.execute_statement(**self._prepare_execute_args("SHOW ERRORS"))
-            err_info = self._render_response(err_res)["records"][-1]
-            return DatabaseError(MySQLErrorCodes(err_info[1]), err_info[2])
-        except self._client.exceptions.BadRequestException:
-            return DatabaseError(original_error)
+            if error_msg.startswith("Database error code"):  # MySQL error
+                code, msg = (s.split(": ", 1)[1] for s in error_msg.split(". ", 1))
+                return DatabaseError(MySQLErrorCodes(int(code)), msg)
+            elif error_msg.startswith("ERROR: "):  # Postgresql error
+                msg, pos, code = postgresql_error_reg.match(error_msg).groups()
+                return DatabaseError(PostgreSQLErrorCodes(code), msg, pos)
+        except Exception:
+            pass
+        return DatabaseError(origin_error)
 
     def execute(self, operation, parameters=None):
         self._current_response, self._iterator, self._paging_state = None, None, None
         execute_statement_args = dict(self._prepare_execute_args(operation),
                                       includeResultMetadata=True)
+        if self._continue_after_timeout is not None:
+            execute_statement_args["continueAfterTimeout"] = self._continue_after_timeout
         if parameters:
             execute_statement_args["parameters"] = self._format_parameter_set(parameters)
         logger.debug("execute %s", reprlib.repr(operation.strip()))
@@ -226,7 +243,7 @@ class AuroraDataAPICursor:
         except self._client.exceptions.BadRequestException as e:
             if "Please paginate your query" in str(e):
                 self._start_paginated_query(execute_statement_args)
-            elif "Database response exceeded size limit" in str(e):
+            elif "Database returned more than the allowed response size limit" in str(e):
                 self._start_paginated_query(execute_statement_args, records_per_page=max(1, self.arraysize // 2))
             else:
                 raise self._get_database_error(e) from e
@@ -283,10 +300,14 @@ class AuroraDataAPICursor:
             if col_desc and col_desc.type_code in self._data_api_type_hint_map:
                 if col_desc.type_code == Decimal:
                     scalar_value = Decimal(scalar_value)
+                elif col_desc.type_code == "json":
+                    scalar_value = json.loads(scalar_value)
+                elif col_desc.type_code == UUID:
+                    scalar_value = UUID(scalar_value)
                 else:
                     try:
                         scalar_value = col_desc.type_code.fromisoformat(scalar_value)
-                    except AttributeError:  # fromisoformat not supported on Python < 3.7
+                    except (AttributeError, ValueError):  # fromisoformat not supported on Python < 3.7
                         if col_desc.type_code == datetime.date:
                             scalar_value = datetime.datetime.strptime(scalar_value, "%Y-%m-%d").date()
                         elif col_desc.type_code == datetime.time:
@@ -319,7 +340,8 @@ class AuroraDataAPICursor:
                 try:
                     page = self._client.execute_statement(**next_page_args)
                 except self._client.exceptions.BadRequestException as e:
-                    if "Database response exceeded size limit" in str(e) and self._paging_state["records_per_page"] > 1:
+                    cur_rpp = self._paging_state["records_per_page"]
+                    if "Database returned more than the allowed response size limit" in str(e) and cur_rpp > 1:
                         self.scroll(-self._paging_state["records_per_page"])  # Rewind the cursor to read the page again
                         logger.debug("Halving records per page")
                         self._paging_state["records_per_page"] //= 2
@@ -377,6 +399,7 @@ class AuroraDataAPICursor:
 
 
 def connect(aurora_cluster_arn=None, secret_arn=None, rds_data_client=None, database=None, host=None, port=None,
-            username=None, password=None, charset=None):
+            username=None, password=None, charset=None, continue_after_timeout=None):
     return AuroraDataAPIClient(dbname=database, aurora_cluster_arn=aurora_cluster_arn,
-                               secret_arn=secret_arn, rds_data_client=rds_data_client, charset=charset)
+                               secret_arn=secret_arn, rds_data_client=rds_data_client, charset=charset,
+                               continue_after_timeout=continue_after_timeout)
